/*
 *	Copyright (c) 2024 Scott Duensing, scott@kangaroopunch.com
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a copy
 *	of this software and associated documentation files (the "Software"), to deal
 *	in the Software without restriction, including without limitation the rights
 *	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *	copies of the Software, and to permit persons to whom the Software is
 *	furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *	SOFTWARE.
 */


#include "oc_output.h"
#include "state.h"
#include "zscii.h"
#include "memory.h"
#include "interpreter.h"
#include "text.h"


void opcodes_new_line(void) {
	textCharPrint('\n');
}


void opcodes_print(void) {
	__state.pc += zsciiPrint(__state.pc, 0);
}


void opcodes_print_addr(void) {
	zsciiPrint(__state.operands[0], 0);
}


void opcodes_print_char(void) {
	char c = zsciiDecodeChar(__state.operands[0]);
	if (c) textCharPrint(c);
}


void opcodes_print_num(void) {
	textNumPrint(__state.operands[0]);
}


void opcodes_print_obj(void) {
	textObjectPrint(__state.operands[0]);
}


void opcodes_print_paddr(void) {
	zsciiPrint(memoryUnpackAddress(__state.operands[0], MEMORY_PRINT), 0);
}


void opcodes_print_ret(void) {
	__state.pc += zsciiPrint(__state.pc, 0);
	textCharPrint('\n');
	interpreterDoReturn(1);
}


void opcodes_output_stream(void) {
	textOutputStream(__state.operands[0], __state.operands[1]);
}
